@@ -26,6 +26,7 @@
 #include <errno.h>
 
 #include <string>
+#include <complex>
 
 #include "syserr.h"
 
@@ -63,9 +64,14 @@ int tokenWORD(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals);
 int tokenNum(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals);
 int tokenFNum(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals);
 int tokenID(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals);
-int tokenRange(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals);
-int tokenNonRange(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals);
+int tokenRange(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals,
+							List* module);
+int tokenNonRange(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals,
+							List* module);
 int tokenGettoken(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals);
+int tokenSkip(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals,
+							List* module);
+int tokenSkipCR(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals);
 
 std::string	Tokenbuf = "";
 
@@ -191,7 +197,8 @@ int token(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals, List* modu
 
 	int ll = ListLength(goalscar);
 	if (ll < 2) {
-		return -1;
+		syserr("usage : <TOKEN VAR pred> \n");
+		return 0;
 	}
 
 	if (goalscar->Cdr()->Car()->kind() != UNDEF) {
@@ -257,7 +264,8 @@ int tokenCmpDirect(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals, N
 
 	if (nstr->kind() != ATOM) {
 //printf("tokenCmpDirect trace 1\n");
-		return -1;
+		syserr("token is not Atom \n");
+		return 0;
 	}
 
 
@@ -314,7 +322,8 @@ int tokenC(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
 	
 	int ll = ListLength(goalscar);
 	if ((ll != 2) && (ll != 1)) {
-		return -1;
+		syserr("usage : <C [VAR]> \n");
+		return 0;
 	}
 
 	int c = fgetc(cx->ioin);
@@ -342,7 +351,7 @@ int tokenC(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
 	}
 	
 	if (goalscar->Cdr()->Car()->kind() != UNDEF) {
-		syserr("usage : <C VAR> \n");
+		syserr("usage : <C [VAR]> \n");
 		return 0;
 	}
 
@@ -361,7 +370,8 @@ int tokenEof(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
 
 	int ll = ListLength(goalscar);
 	if (ll != 1) {
-		return -1;
+		syserr("usage : <EOF> \n");
+		return 0;
 	}
 
 	int c = fgetc(cx->ioin);
@@ -385,7 +395,8 @@ int tokenN(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
 
 	int ll = ListLength(goalscar);
 	if ((ll != 2) && (ll != 1)) {
-		return -1;
+		syserr("usage : <N [VAR]> \n");
+		return 0;
 	}
 
 	int c = fgetc(cx->ioin);
@@ -397,7 +408,7 @@ int tokenN(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
 		ungetc(c, cx->ioin);
 		return -1;
 	}
-		
+
 	Tokenbuf = (char)c;
 
 	if (ll == 1) {
@@ -405,7 +416,7 @@ int tokenN(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
 	}
 
 	if (goalscar->Cdr()->Car()->kind() != UNDEF) {
-		syserr("usage : <N VAR> \n");
+		syserr("usage : <N [VAR]> \n");
 		return 0;
 	}
 
@@ -424,7 +435,8 @@ int tokenA(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
 
 	int ll = ListLength(goalscar);
 	if ((ll != 2) && (ll != 1)) {
-		return -1;
+		syserr("usage : <A [VAR]> \n");
+		return 0;
 	}
 
 
@@ -447,7 +459,7 @@ int tokenA(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
 	Tokenbuf = (char)c;
 
 	if (goalscar->Cdr()->Car()->kind() != UNDEF) {
-		syserr("usage : <A VAR> \n");
+		syserr("usage : <A [VAR]> \n");
 		return 0;
 	}
 
@@ -466,7 +478,8 @@ int tokenAN(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
 
 	int ll = ListLength(goalscar);
 	if ((ll != 2) && (ll != 1)) {
-		return -1;
+		syserr("usage : <AN [VAR]> \n");
+		return 0;
 	}
 
 
@@ -490,7 +503,7 @@ int tokenAN(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
 	}
 
 	if (goalscar->Cdr()->Car()->kind() != UNDEF) {
-		syserr("usage : <AN VAR> \n");
+		syserr("usage : <AN [VAR]> \n");
 		return 0;
 	}
 
@@ -510,7 +523,8 @@ int tokenLineHead(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
 
 	int ll = ListLength(goalscar);
 	if (ll != 1) {
-		return -1;
+		syserr("usage : <^> \n");
+		return 0;
 	}
 
 
@@ -538,7 +552,8 @@ int tokenLineEnd(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
 
 	int ll = ListLength(goalscar);
 	if (ll != 1) {
-		return -1;
+		syserr("usage : <$> \n");
+		return 0;
 	}
 
 	int c = fgetc(cx->ioin);
@@ -565,7 +580,8 @@ int tokenCR(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
 
 	int ll = ListLength(goalscar);
 	if (ll != 1) {
-		return -1;
+		syserr("usage : <CR> \n");
+		return 0;
 	}
 
 	int c = fgetc(cx->ioin);
@@ -588,7 +604,8 @@ int tokenCNTL(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
 
 	int ll = ListLength(goalscar);
 	if ((ll != 2) && (ll != 1)) {
-		return -1;
+		syserr("usage : <CNTL [VAR]> \n");
+		return 0;
 	}
 
 	int c = fgetc(cx->ioin);
@@ -604,7 +621,7 @@ int tokenCNTL(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
 	}
 
 	if (goalscar->Cdr()->Car()->kind() != UNDEF) {
-		syserr("usage : <CNTL VAR> \n");
+		syserr("usage : <CNTL [VAR]> \n");
 		return 0;
 	}
 
@@ -625,7 +642,8 @@ int tokenSpace(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
 	int	i, n;
 	int ll = ListLength(goalscar);
 	if (ll != 1) {
-		return -1;
+		syserr("usage : <SPACE> \n");
+		return 0;
 	}
 
 	int c = fgetc(cx->ioin);
@@ -648,7 +666,8 @@ int tokenPunct(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
 
 	int ll = ListLength(goalscar);
 	if ((ll != 2) && (ll != 1)) {
-		return -1;
+		syserr("usage : <PUNCT [VAR]> \n");
+		return 0;
 	}
 
 	int c = fgetc(cx->ioin);
@@ -664,7 +683,7 @@ int tokenPunct(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
 	}
 
 	if (goalscar->Cdr()->Car()->kind() != UNDEF) {
-		syserr("usage : <PUNCT VAR> \n");
+		syserr("usage : <PUNCT [VAR]> \n");
 		return 0;
 	}
 
@@ -685,11 +704,12 @@ int tokenStrings(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
 
 	int ll = ListLength(goalscar);
 	if ((ll != 2) && (ll != 1)) {
-		return -1;
+		syserr("usage : <STRINGS [VAR]> \n");
+		return 0;
 	}
 
 	if ((ll == 2) && (goalscar->Cdr()->Car()->kind() != UNDEF)) {
-		syserr("usage : <STRINGS VAR> \n");
+		syserr("usage : <STRINGS [VAR]> \n");
 		return 0;
 	}
 
@@ -709,6 +729,16 @@ int tokenStrings(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
 
 	for (;;) {
 		c = fgetc(cx->ioin);
+		int n = CharLen(c);
+		if (n > 1) {
+			Tokenbuf = Tokenbuf + (char)(c & 0xff);
+			do {
+				c = fgetc(cx->ioin);
+				Tokenbuf = Tokenbuf + (char)(c & 0xff);
+				n--;
+			} while(n > 1);
+			continue;
+		}
 
 		if (c == qt) {
 			Node* env=Nil->Cons(Nil);
@@ -740,11 +770,12 @@ int tokenWORD(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
 	
 	int ll = ListLength(goalscar);
 	if ((ll != 2) && (ll != 1)) {
-		return -1;
+		syserr("usage : <WORD [VAR]> \n");
+		return 0;
 	}
 
 	if ((ll == 2) && (goalscar->Cdr()->Car()->kind() != UNDEF)) {
-		syserr("usage : <WORD VAR> \n");
+		syserr("usage : <WORD [VAR]> \n");
 		return 0;
 	}
 
@@ -859,11 +890,12 @@ int tokenWildcard(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
 
 	int ll = ListLength(goalscar);
 	if ((ll != 2) && (ll != 1)) {
-		return -1;
+		syserr("usage : <* [VAR]> \n");
+		return 0;
 	}
 
 	if ((ll == 2) && (goalscar->Cdr()->Car()->kind() != UNDEF)) {
-		syserr("usage : <* VAR> \n");
+		syserr("usage : <* [VAR]> \n");
 		return 0;
 	}
 
@@ -958,11 +990,12 @@ int tokenNum(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
 	
 	int ll = ListLength(goalscar);
 	if ((ll != 2) && (ll != 1)) {
-		return -1;
+		syserr("usage : <NUM [VAR]> \n");
+		return 0;
 	}
 
 	if ((ll == 2) && (goalscar->Cdr()->Car()->kind() != UNDEF)) {
-		syserr("usage : <NUM VAR> \n");
+		syserr("usage : <NUM [VAR]> \n");
 		return 0;
 	}
 
@@ -1004,11 +1037,12 @@ int tokenFNum(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
 
 	int ll = ListLength(goalscar);
 	if ((ll != 2) && (ll != 1)) {
-		return -1;
+		syserr("usage : <FNUM [VAR]> \n");
+		return 0;
 	}
 
 	if ((ll == 2) && (goalscar->Cdr()->Car()->kind() != UNDEF)) {
-		syserr("usage : <FNUM VAR> \n");
+		syserr("usage : <FNUM [VAR]> \n");
 		return 0;
 	}
 
@@ -1098,11 +1132,12 @@ int tokenID(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
 	
 	int ll = ListLength(goalscar);
 	if ((ll != 2) && (ll != 1)) {
-		return -1;
+		syserr("usage : <ID [VAR]> \n");
+		return 0;
 	}
 
 	if ((ll == 2) && (goalscar->Cdr()->Car()->kind() != UNDEF)) {
-		syserr("usage : <ID VAR> \n");
+		syserr("usage : <ID [VAR]> \n");
 		return 0;
 	}
 
@@ -1186,12 +1221,14 @@ int tokenID(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
 	return -1;
 }
 
-int tokenRangeSub(Context* cx, Node* goalscar, int flg)
+int tokenRangeSub(Context* cx, Node* goalscar, int flg, List* module)
 {
 	int		c, i, n;
 	long long	cl, cl1, cl2;
 	std::string	s;
 	
+	long tp = ftell(cx->ioin);
+
 	int ll = ListLength(goalscar);
 	if (ll != 4) {
 		return -1;
@@ -1208,10 +1245,28 @@ int tokenRangeSub(Context* cx, Node* goalscar, int flg)
 
 
 	// 2nd args
-	if (goalscar->Cdr()->Cdr()->Car()->kind() != ATOM) {
+	Node* arg2 = goalscar->Cdr()->Cdr()->Car();
+	int rn;
+
+	if ((rn = FuncArg(cx, arg2, goalscar, module)) <= 0) {
+		if (flg) {
+			syserr("RANGE: failed in the evaluation of the argument. \n");
+		} else {
+			syserr("NONRANGE: failed in the evaluation of the argument. \n");
+		}
+		return 0;
+	}
+
+	if (arg2->kind() != ATOM) {
+		if (flg) {
+			syserr("usage : <RANGE char char> \n");
+		} else {
+			syserr("usage : <NONRANGE char char> \n");
+		}
 		return -1;
 	}
-	((Atom*)(goalscar->Cdr()->Cdr()->Car()))->toString(s);
+
+	((Atom*)arg2)->toString(s);
 	c = s[0];
 	n = CharLen(c);
 	cl1 = c & 0xff;
@@ -1223,10 +1278,27 @@ int tokenRangeSub(Context* cx, Node* goalscar, int flg)
 //printf("RANGE args2 %llx \n", cl1);
 
 	// 3rd args
-	if (goalscar->Cdr()->Cdr()->Cdr()->Car()->kind() != ATOM) {
+	Node* arg3 = goalscar->Cdr()->Cdr()->Cdr()->Car();
+
+	if ((rn = FuncArg(cx, arg3, goalscar, module)) <= 0) {
+		if (flg) {
+			syserr("RANGE: failed in the evaluation of the argument. \n");
+		} else {
+			syserr("NONRANGE: failed in the evaluation of the argument. \n");
+		}
+		return 0;
+	}
+
+	if (arg3->kind() != ATOM) {
+		if (flg) {
+			syserr("usage : <RANGE char char> \n");
+		} else {
+			syserr("usage : <NONRANGE char char> \n");
+		}
 		return -1;
 	}
-	((Atom*)(goalscar->Cdr()->Cdr()->Cdr()->Car()))->toString(s);
+	((Atom*)arg3)->toString(s);
+
 	c = s[0];
 	n = CharLen(c);
 	cl2 = c & 0xff;
@@ -1238,6 +1310,11 @@ int tokenRangeSub(Context* cx, Node* goalscar, int flg)
 //printf("RANGE args3 %llx \n", cl2);
 
 	if (cl2 < cl1) {
+		if (flg) {
+			syserr("usage : <RANGE char char> \n");
+		} else {
+			syserr("usage : <NONRANGE char char> \n");
+		}
 		return -1;
 	}	
 
@@ -1268,30 +1345,31 @@ int tokenRangeSub(Context* cx, Node* goalscar, int flg)
 		}
 	}
 
-	for (i = n-1; i > 0; i++) {
-		ungetc(Tokenbuf[i], cx->ioin);
-	}
+	fseek(cx->ioin, tp, 0);
+
 	return -1;
 }
 
-int tokenRange(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
+int tokenRange(Context* cx, Node* goalscar, Node* &goalscdr, 
+					Node* &goals, List* module)
 {
 	if (cx->ioin == stdin) {
 		syserr("File is not opened");
 		return 0;
 	}
 
-	return tokenRangeSub(cx, goalscar, 1);
+	return tokenRangeSub(cx, goalscar, 1, module);
 }
 
-int tokenNonRange(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
+int tokenNonRange(Context* cx, Node* goalscar, Node* &goalscdr, 
+					Node* &goals, List* module)
 {
 	if (cx->ioin == stdin) {
 		syserr("File is not opened");
 		return 0;
 	}
 
-	return tokenRangeSub(cx, goalscar, 0);
+	return tokenRangeSub(cx, goalscar, 0, module);
 }
 
 int tokenGettoken(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
@@ -1303,7 +1381,8 @@ int tokenGettoken(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
 
 	int ll = ListLength(goalscar);
 	if (ll != 2) {
-		return -1;
+		syserr("usage : <GETTOKEN VAR>\n");
+		return 0;
 	}
 
 	if (goalscar->Cdr()->Car()->kind() != UNDEF) {
@@ -1315,3 +1394,108 @@ int tokenGettoken(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
 	return TokenVar(cx, goalscar);
 }
 
+int tokenSkip(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals,	
+							List* module)
+{
+	int c;
+	
+	if (cx->ioin == stdin) {
+		syserr("File is not opened");
+		return 0;
+	}
+
+	int ll = ListLength(goalscar);
+	if (ll != 2) {
+		syserr("usage : <SKIP STR>\n");
+		return 0;
+	}
+
+	Node* ag = goalscar->Cdr()->Car();
+	int rn;
+
+	if ((rn = FuncArg(cx, ag, goalscar, module)) <= 0) {
+		syserr("SKIP: failed in the evaluation of the argument. \n");
+		return 0;
+	}
+
+	if (ag->kind() != ATOM) {
+		syserr("usage : <SKIP STR>\n");
+		return 0;
+	}
+
+	std::string s, subs;
+
+	((Atom*)ag)->toString(s);
+
+	int i, n;
+	for (;;) {
+		c = fgetc(cx->ioin);
+		if (c == EOF) {
+			return 0;
+		}
+		n = CharLen(c);
+		if ((char)(s[0] & 0xff) != (char)(c & 0xff)) {
+			for (i = 1; i < n; i++) {
+				c=fgetc(cx->ioin);
+			}
+		} else {
+			subs = "";
+			subs = subs + (char)c;
+			for (i = 1; i < n; i++) {
+				c=fgetc(cx->ioin);
+				subs = subs + (char)c;
+			}
+			long fpo = ftell(cx->ioin);
+			
+			for (i = n; i < s.length(); i++) {
+				c = fgetc(cx->ioin);
+				subs = subs + (char)c;
+			}
+			if (s == subs) {
+				break;
+			}
+			fseek(cx->ioin, fpo, 0);
+		} 
+	}
+		
+		
+	return 1;
+}
+
+int tokenSkipCR(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
+{
+	int c;
+	
+	if (cx->ioin == stdin) {
+		syserr("File is not opened");
+		return 0;
+	}
+
+	int ll = ListLength(goalscar);
+	if (ll != 1) {
+		syserr("usage : <SKIPCR>\n");
+		return 0;
+	}
+
+	if (goalscar->Cdr()->Car()->kind() != ATOM) {
+		syserr("usage : <SKIPCR>\n");
+		return 0;
+	}
+
+	int i, n;
+	for (;;) {
+		c = fgetc(cx->ioin);
+		if (c == EOF) {
+			return 1;
+		}
+		if (c == '\n') {
+			break;
+		} 
+	}
+		
+		
+	return 1;
+}
+
+
+
